#include <iomanip>
#include <time.h>
#include <ctype.h>
#include <unistd.h>
#include <math.h>
#include <stdlib.h>     // for (s)random
#include <sys/stat.h>   // for mkdir
#include <sys/types.h>

#include <set>
#include <iostream>

#include "imms.h"
#include "strmanip.h"
#include "md5digest.h"

using std::cerr;

Imms::Imms()
    : threshold(INITIAL_RATING), timehold(24*60),
    last_timehold(timehold), last_threshold(threshold)
{
    verbose = last_skipped = last_jumped = did_check = false;

    srandom(time(0));

    string dotimms = getenv("HOME");
    mkdir(dotimms.append("/.imms").c_str(), 0700);

    fout.open(dotimms.append("/imms.log").c_str(),
            ofstream::out | ofstream::app);

    sqldb.reset(new SqlDb());
}

void Imms::setup(const char* _email, int _quality, bool _verbose)
{
    email = _email;
    quality = _quality;
    verbose = _verbose;
    time_t t = time(0);
    fout << endl << endl << ctime(&t) << "User: '" << email << "'"
        << " [" << (quality > 0 ? "+" : "-") << (quality > 1 ? "+" : "-")
        << (quality > 2 ? "+" : "-") << "] " << (verbose ? "(*)" : "");
    fout.flush();
}

void Imms::start_song(const string &path)
{
    if (!did_check)
        fetch_song_info(path);

    print_song_info(path);

    update_thresholds(path);
}

void Imms::print_song_info(const string &path)
{
    fout << endl << string(TERM_WIDTH - 15, '-') << endl;

    if (path.length() > TERM_WIDTH - 8)
        fout << "[Path: ..." << path.substr(path.length() - TERM_WIDTH + 11,
                TERM_WIDTH - 11) << "]" << endl;
    else
        fout << "[Path: " << path << "]" << endl;

    fout << "  [Rating: " << rating << "/" << setprecision(3) <<
        last_threshold << setprecision(5) << "/";

    if (!last_jumped && rating < last_threshold)
    {
        float total = pow(2, (last_threshold - rating)/EXPONENT_STEP) + BASE_BIAS;
        int twodigit = ROUND(100 / total);
        fout << (float)twodigit / 100 << "] ";
    }
    else
        fout << "-] ";

    fout << "[Time: " << strtime(last_time) << "/" << setprecision(3)
        << strtime(last_timehold) << setprecision(5) << "] ";

    if (!identified) 
        fout << "[Unknown] ";

    fout.flush();
}

bool Imms::check_song(const string &path)
{
    if (access(path.c_str(), R_OK))
        return false;

    fetch_song_info(path);

    bool status = true;

    if (!last_jumped && rating < threshold)
    {
        // Figure out how far below the threshold this song falls
        float total = pow(2, (threshold - rating)/EXPONENT_STEP) + BASE_BIAS;

        if ((random() % 100) * total / 100 > 1) 
            status = false;
    }

    last_threshold = threshold;
    float w = rating > threshold ? THRESHOLD_GROW + quality * QUALITY_BIAS 
        : THRESHOLD_DECAY;
    threshold = threshold * (1 - w) + rating * w;

    last_timehold = timehold;
    if (rating > threshold)
    {
        timehold = ROUND(timehold * (1 - TIMEHOLD_STEP)
                + last_time * TIMEHOLD_STEP);
    }

    if (last_time + 1 < timehold && !last_jumped)
    {
        status = false;
    }

    did_check = true;

    if (!status && verbose)
    {
        print_song_info(path);
        fout << "[Auto] ";
    }

    return status;
}

void Imms::update_thresholds(const string &path)
{
    if (!did_check)
    {
        float w = rating > threshold ? THRESHOLD_GROW + quality * QUALITY_BIAS
            : THRESHOLD_DECAY;
        threshold = threshold * (1 - w) + rating * w;

        if (rating > threshold)
        {
            timehold = ROUND(timehold * (1 - TIMEHOLD_STEP)
                    + last_time * TIMEHOLD_STEP);
        }
    }

    sqldb->set_last(time(0));
}

void Imms::end_song(bool at_the_end, bool jumped, bool bad)
{
    int mod;

    if (at_the_end)
    {
        if (last_jumped)
            mod = JUMPED_TO_FINNISHED;
        else if (last_skipped)
            mod = FIRST_NON_SKIP_RATE;
        else
            mod = CONS_NON_SKIP_RATE;
        last_skipped = jumped = false;
    }
    else
    {
        if (last_jumped && !jumped)
            mod = JUMPED_TO_SKIPPED;
        else if (jumped)
            mod = JUMPED_FROM;
        else if (last_skipped)
            mod = CONS_SKIP_RATE;
        else
            mod = FIRST_SKIP_RATE;
        last_skipped = true;
    }

    if (bad)
        mod = 0;
    else if (rating > 200)
        mod /= 4;
    else if (rating > 150)
        mod /= 2;

    if (last_time < last_timehold)
        mod /= 2;

    fout << (jumped ? "[Jumped] " : "");
    fout << (!jumped && last_skipped ? "[Skipped] " : "");
    fout << "[Delta " << setiosflags(ios::showpos) << mod <<
        resetiosflags (ios::showpos) << "] ";
    fout << (unrated ? "[New] " : "");

    last_jumped = jumped;
    did_check = false;

    sqldb->set_rating(rating + mod);
}

void Imms::fetch_song_info(const string &path)
{
    if (access(path.c_str(), R_OK))
        return;

    struct stat statbuf;
    stat(path.c_str(), &statbuf);

    if (sqldb->identify(path, statbuf.st_mtime) < 0)
    {
        if (sqldb->identify(path, statbuf.st_mtime,
                Md5Digest::digest_file(path)) < 0)
            return;
    }

    songinfo.link(path);
    rating = sqldb->get_rating();

    unrated = false;
    if (rating < 0)
    {
        unrated = true;
        rating = songinfo.get_rating(email);
    }

    artist = sqldb->get_artist();
    title = sqldb->get_title();

    if (artist != "" && title != "")
    {
        identified = true;
    }
    else
    {
        parse_song_info(path);
        if (identified)
        {
            sqldb->set_title(title);
        }
    }

#ifdef DEBUG
    cerr << "========================================" << endl;
    cerr << "path:\t" << path << endl;
    cerr << "artist:\t" << artist << (identified ? " *" : "") << endl;
    cerr << "title:\t" << title << (identified ? " *" : "") << endl;
#endif

    last_time = (time(0) - sqldb->get_last()) / 60;

    if (last_time > MAX_TIME)
        last_time = MAX_TIME;
}

void Imms::parse_song_info(const string &path)
{
    identified = false;
    bool artist_confirmed = false;

    string tag_artist = artist = string_normalize(songinfo.get_artist());
    title = string_normalize(songinfo.get_title());

    int last_slash = path.find_last_of("/");
    string only_path = path.substr(0, last_slash);
    string filename = path.substr(last_slash + 1,
            path.length() - last_slash - 5);

    set<string> file_parts;
    imms_magic_parse(file_parts, filename);

    set<string> path_parts(file_parts);
    imms_magic_parse(path_parts, only_path);

    // See if we can recognize any of the path parts
    for (set<string>::iterator i = path_parts.begin();
            !artist_confirmed && i != path_parts.end(); i++)
    {
        if (artist_confirmed =
                (sqldb->check_artist(*i) || string_like(*i, artist)))
        {
            artist = *i;
        }
    }

    if (tag_artist != "" && tag_artist != artist
            && !string_like(artist, tag_artist))
    {
        if (sqldb->check_artist(tag_artist))
        {
            artist = tag_artist;
            artist_confirmed = true;
        }
    }

    // If we still don't know what to do - give up
    if (!artist_confirmed)
        return;

    // By this point the artist had to have been confirmed
    sqldb->set_artist(artist);

    // If we know the title and were only missing the artist
    if (title != "")
        identified = sqldb->check_title(title);

    // Can recognize the title by matching filename parts to the database?
    for (set<string>::iterator i = file_parts.begin();
            !identified && i != file_parts.end(); i++)
    {
        if (identified = (sqldb->check_title(*i) || string_like(*i, title)))
            title = *i;
    }

    if (identified)
        return;

    file_parts.erase(artist);

    if (file_parts.size() != 1 ||
            file_parts.begin()->find(artist) != string::npos)
        return;

    if (title.find("mix") == string::npos
            && title.find("version") == string::npos)
    {
        title = *file_parts.begin();
        identified = true;
        return;
    }

    for (set<string>::iterator i = path_parts.begin();
            !identified && i != path_parts.end(); i++)
    {
        if (artist == *i)
            continue;
        if (identified = sqldb->check_title(*i))
            title = *i;
    }
}
