/*
 *      The XMMS plugin portion of IMMS
 *      GUI implementation is based on XMMS's song_change plugin
 */


#include <string>
#include <iostream>

#include <xmms/plugin.h>
#include <xmms/xmmsctrl.h> 
#include <xmms/configfile.h> 

#include <gtk/gtk.h>

#include "imms.h"
#include "immsconf.h"

#define DEFAULT_EMAIL "default@imms.org"

using std::string;
using std::cerr;

// Local vars

static Imms *imms = NULL;
static unsigned int time_left = 1000;
static int poll_tag = 0, last_plpos = -1, cur_plpos;
static int was_non_zero = 0;

static string cur_path = "", last_path = "";

static enum
{
    NOT_BUSY    = 0,
    BUSY        = 1,
    FIND_NEXT   = 2
} state;

// GUI junk

static GtkWidget *configure_win = NULL, *configure_vbox;
static GtkWidget *email_entry, *verbose_button;
static GtkObject *quality_adjustment;

// Conf vars

gchar *ch_email = NULL;
int quality = 0, verbose = 0;

// C interface

extern "C"
{
    static void init(void);
    static void cleanup(void);
    static void configure(void);
    static gint poll_func(gpointer);
    GeneralPlugin *get_gplugin_info(void);
}

GeneralPlugin imms_gp =
{
    NULL,                   /* handle */
    NULL,                   /* filename */
    -1,                     /* xmms_session */
    NULL,                   /* Description */
    init,
    NULL,
    configure,
    cleanup,
};

int &xmms_session = imms_gp.xmms_session;

string my_xmms_get_path(int plpos)
{
    gchar *tmp = 0;
    while (!tmp)
       tmp = xmms_remote_get_playlist_file(xmms_session, plpos);
    string result = tmp;
    g_free(tmp);
    return result;
}

GeneralPlugin *get_gplugin_info(void)
{
    imms_gp.description = PACKAGE_STRING;
    return &imms_gp;
}

static void read_config(void)
{
    ConfigFile *cfgfile;

    g_free(ch_email);
    ch_email = NULL;

    if ((cfgfile = xmms_cfg_open_default_file()) != NULL)
    {
        xmms_cfg_read_string(cfgfile, "imms", "email", &ch_email);
        xmms_cfg_read_int(cfgfile, "imms", "quality", &quality);
        xmms_cfg_read_int(cfgfile, "imms", "verbose", &verbose);
        xmms_cfg_free(cfgfile);
    }
}

static void init(void)
{
    read_config();

    if (!imms)
    {
        imms = new Imms();
        imms->setup(ch_email ? ch_email : DEFAULT_EMAIL, quality, verbose);
    }

    state = NOT_BUSY;
    poll_tag = gtk_timeout_add(20, poll_func, NULL);
}

static void cleanup(void)
{
    delete imms;
    imms = NULL;

    if (poll_tag)
        gtk_timeout_remove(poll_tag);

    poll_tag = 0;
}

static gint poll_func(gpointer)
{
    switch (state)
    {
        case BUSY:
            return TRUE;

        case FIND_NEXT:
            state = BUSY;

            {
                int len = xmms_remote_get_playlist_length(xmms_session);
                bool forced = time_left && (last_plpos + 1 != cur_plpos) &&
                    (cur_plpos != 0 || last_plpos != len);

                if (time_left)
                    cerr << time_left << endl;

                if (last_path != "")
                    imms->end_song(!time_left, forced, was_non_zero);

                if (!forced && len > 2)
                {
                    do {
                        cur_plpos = random() % len;
                        cur_path = my_xmms_get_path(cur_plpos);
                    }
                    while (!imms->check_song(cur_path));

                    xmms_remote_set_playlist_pos(xmms_session, cur_plpos);
                }
            }

            imms->start_song(cur_path);

            last_path = cur_path;
            was_non_zero = 25;

            xmms_remote_play(xmms_session);

            state = NOT_BUSY;
            return TRUE;

        case NOT_BUSY:

            if (!xmms_remote_is_playing(xmms_session))
                return TRUE;

            state = BUSY;

            if (xmms_remote_is_shuffle(xmms_session))
                xmms_remote_toggle_shuffle(xmms_session);

            cur_plpos = xmms_remote_get_playlist_pos(xmms_session);

            time_left = (xmms_remote_get_playlist_time(xmms_session, cur_plpos)
                    - xmms_remote_get_output_time(xmms_session)) >> 9;

            cur_path = my_xmms_get_path(cur_plpos);

            if (last_path != cur_path || (!was_non_zero && !time_left))
            {
                xmms_remote_stop(xmms_session);

                state = FIND_NEXT;
                return TRUE;
            }

            last_plpos = cur_plpos;
            if (was_non_zero && time_left)
                was_non_zero--;

            state = NOT_BUSY;
            return TRUE;
    }

    return TRUE;
}

static void configure_ok_cb(gpointer data)
{
    ConfigFile *cfgfile;

    gchar *email = gtk_entry_get_text(GTK_ENTRY(email_entry));
    quality = ROUND(GTK_ADJUSTMENT(quality_adjustment)->value);
    verbose = !!GTK_TOGGLE_BUTTON(verbose_button)->active;

    cfgfile = xmms_cfg_open_default_file();
    xmms_cfg_write_string(cfgfile, "imms", "email", email);
    xmms_cfg_write_int(cfgfile, "imms", "quality", quality);
    xmms_cfg_write_int(cfgfile, "imms", "verbose", verbose);
    xmms_cfg_write_default_file(cfgfile);
    xmms_cfg_free(cfgfile);

    g_free(ch_email);
    ch_email = g_strdup(email);

    if (imms)
    {
        imms->setup(ch_email, quality, verbose);   
    }
    gtk_widget_destroy(configure_win);
}  

static void configure(void)
{
    if (configure_win)
        return;

    read_config();

    configure_win = gtk_window_new(GTK_WINDOW_DIALOG);
    gtk_signal_connect(GTK_OBJECT(configure_win), "destroy",
            GTK_SIGNAL_FUNC(gtk_widget_destroyed), &configure_win);
    gtk_window_set_title(GTK_WINDOW(configure_win), "IMMS Configuration");

    gtk_container_set_border_width(GTK_CONTAINER(configure_win), 10);

    configure_vbox = gtk_vbox_new(FALSE, 10);
    gtk_container_add(GTK_CONTAINER(configure_win), configure_vbox);

    // Email
    GtkWidget *email_hbox, *email_label;
    GtkWidget *email_desc, *email_vbox, *email_frame; 
   
    email_frame = gtk_frame_new("Email");
    gtk_box_pack_start(GTK_BOX(configure_vbox), email_frame, FALSE, FALSE, 0);
    email_vbox = gtk_vbox_new(FALSE, 10);
    gtk_container_set_border_width(GTK_CONTAINER(email_vbox), 5);
    gtk_container_add(GTK_CONTAINER(email_frame), email_vbox);
    gtk_widget_show(email_frame);
    gtk_widget_show(email_vbox);

    email_desc = gtk_label_new("The email to identify the user by");
    gtk_label_set_justify(GTK_LABEL(email_desc), GTK_JUSTIFY_LEFT);
    gtk_misc_set_alignment(GTK_MISC(email_desc), 0, 0.5);
    gtk_box_pack_start(GTK_BOX(email_vbox), email_desc, FALSE, FALSE, 0);
    gtk_label_set_line_wrap(GTK_LABEL(email_desc), TRUE);
    gtk_widget_show(email_desc);

    email_hbox = gtk_hbox_new(FALSE, 5);
    gtk_box_pack_start(GTK_BOX(email_vbox), email_hbox, FALSE, FALSE, 0);

    email_label = gtk_label_new("Email:");
    gtk_box_pack_start(GTK_BOX(email_hbox), email_label, FALSE, FALSE, 0);
    gtk_widget_show(email_label);

    email_entry = gtk_entry_new();
    if (ch_email)
        gtk_entry_set_text(GTK_ENTRY(email_entry), ch_email);
    else
        gtk_entry_set_text(GTK_ENTRY(email_entry), DEFAULT_EMAIL);
    gtk_widget_set_usize(email_entry, 200, -1);
    gtk_box_pack_start(GTK_BOX(email_hbox), email_entry, TRUE, TRUE, 0);
    gtk_widget_show(email_entry);
    gtk_widget_show(email_hbox);
    
    // Quality
    GtkWidget *qual_frame, *qual_vbox, *qual_desc;
    GtkWidget *qual_scroll, *qual_hbox, *qual_label, *qual_label_2;

    qual_frame = gtk_frame_new("Quality");
    gtk_box_pack_start(GTK_BOX(configure_vbox), qual_frame, FALSE, FALSE, 0);
    qual_vbox = gtk_vbox_new(FALSE, 10);
    gtk_container_set_border_width(GTK_CONTAINER(qual_vbox), 5);
    gtk_container_add(GTK_CONTAINER(qual_frame), qual_vbox);
    gtk_widget_show(qual_frame);
    gtk_widget_show(qual_vbox);

    qual_desc = gtk_label_new("Learning vs. average song quality tradeoff. "
           "Use low values for untrained databases.");
    gtk_label_set_justify(GTK_LABEL(qual_desc), GTK_JUSTIFY_LEFT);
    gtk_misc_set_alignment(GTK_MISC(qual_desc), 0, 0.5);
    gtk_box_pack_start(GTK_BOX(qual_vbox), qual_desc, FALSE, FALSE, 0);
    gtk_label_set_line_wrap(GTK_LABEL(qual_desc), TRUE);
    gtk_widget_show(qual_desc);

    qual_hbox = gtk_hbox_new(FALSE, 5);
    gtk_box_pack_start(GTK_BOX(qual_vbox), qual_hbox, FALSE, FALSE, 0);

    qual_label = gtk_label_new("Learn");
    gtk_box_pack_start(GTK_BOX(qual_hbox), qual_label, FALSE, FALSE, 0);
    gtk_widget_show(qual_label);

    quality_adjustment = gtk_adjustment_new(quality, 0, 5, 1, 1, 1);
    qual_scroll = gtk_hscrollbar_new(GTK_ADJUSTMENT(quality_adjustment));

    gtk_widget_set_usize(qual_scroll, 200, -1);
    gtk_box_pack_start(GTK_BOX(qual_hbox), qual_scroll, TRUE, TRUE, 0);
    gtk_widget_show(qual_scroll);

    qual_label_2 = gtk_label_new("Quality");
    gtk_box_pack_start(GTK_BOX(qual_hbox), qual_label_2, FALSE, TRUE, 0);
    gtk_widget_show(qual_label_2);
    gtk_widget_show(qual_hbox);

    // Verbosity
    GtkWidget *verb_hbox;
    GtkWidget *verb_vbox, *verb_frame; 
   
    verb_frame = gtk_frame_new("Verbsity");
    gtk_box_pack_start(GTK_BOX(configure_vbox), verb_frame, FALSE, FALSE, 0);
    verb_vbox = gtk_vbox_new(FALSE, 10);
    gtk_container_set_border_width(GTK_CONTAINER(verb_vbox), 5);
    gtk_container_add(GTK_CONTAINER(verb_frame), verb_vbox);
    gtk_widget_show(verb_frame);
    gtk_widget_show(verb_vbox);

    verb_hbox = gtk_hbox_new(FALSE, 5);
    gtk_box_pack_start(GTK_BOX(verb_vbox), verb_hbox, FALSE, FALSE, 0);

    verbose_button =
        gtk_check_button_new_with_label("Include skipped songs in the log");
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(verbose_button), verbose);
    gtk_box_pack_start(GTK_BOX(verb_hbox), verbose_button, FALSE, FALSE, 0);

    gtk_widget_show(verbose_button);
    gtk_widget_show(verb_hbox);


    // Buttons
    GtkWidget *configure_bbox, *configure_ok, *configure_cancel;

    configure_bbox = gtk_hbutton_box_new();
    gtk_button_box_set_layout(GTK_BUTTON_BOX(configure_bbox), GTK_BUTTONBOX_END);
    gtk_button_box_set_spacing(GTK_BUTTON_BOX(configure_bbox), 5);
    gtk_box_pack_start(GTK_BOX(configure_vbox), configure_bbox, FALSE, FALSE, 0);

    configure_ok = gtk_button_new_with_label("Ok");
    gtk_signal_connect(GTK_OBJECT(configure_ok), "clicked",
            GTK_SIGNAL_FUNC(configure_ok_cb), NULL);
    GTK_WIDGET_SET_FLAGS(configure_ok, GTK_CAN_DEFAULT);
    gtk_box_pack_start(GTK_BOX(configure_bbox), configure_ok, TRUE, TRUE, 0);
    gtk_widget_show(configure_ok);
    gtk_widget_grab_default(configure_ok);

    configure_cancel = gtk_button_new_with_label("Cancel");
    gtk_signal_connect_object(GTK_OBJECT(configure_cancel), "clicked",
            GTK_SIGNAL_FUNC(gtk_widget_destroy), GTK_OBJECT(configure_win));
    GTK_WIDGET_SET_FLAGS(configure_cancel, GTK_CAN_DEFAULT);
    gtk_box_pack_start(GTK_BOX(configure_bbox), configure_cancel, TRUE, TRUE, 0);
    gtk_widget_show(configure_cancel);
    gtk_widget_show(configure_bbox);
    gtk_widget_show(configure_vbox);
    gtk_widget_show(configure_win);
}
